@@ -16,7 +16,9 @@
 #include "sharedialog.h"
 #include "sharee.h"
 #include "sharelinkwidget.h"
+#include "internallinkwidget.h"
 #include "shareusergroupwidget.h"
+#include "passwordinputdialog.h"
 
 #include "sharemanager.h"
 
@@ -25,6 +27,7 @@
 #include "configfile.h"
 #include "theme.h"
 #include "thumbnailjob.h"
+#include "wordlist.h"
 
 #include <QFileInfo>
 #include <QFileIconProvider>
@@ -32,6 +35,21 @@
 #include <QPointer>
 #include <QPushButton>
 #include <QFrame>
+#include <QScrollBar>
+
+namespace {
+QString createRandomPassword()
+{
+    const auto words = OCC::WordList::getRandomWords(10);
+
+    const auto addFirstLetter = [](const QString &current, const QString &next) -> QString {
+        return current + next.at(0);
+    };
+
+    return std::accumulate(std::cbegin(words), std::cend(words), QString(), addFirstLetter);
+}
+}
+
 
 namespace OCC {
 
@@ -42,6 +60,7 @@ ShareDialog::ShareDialog(QPointer<AccountState> accountState,
     const QString &localPath,
     SharePermissions maxSharingPermissions,
     const QByteArray &numericFileId,
+    SyncJournalFileLockInfo filelockState,
     ShareDialogStartPage startPage,
     QWidget *parent)
     : QDialog(parent)
@@ -50,12 +69,9 @@ ShareDialog::ShareDialog(QPointer<AccountState> accountState,
     , _sharePath(sharePath)
     , _localPath(localPath)
     , _maxSharingPermissions(maxSharingPermissions)
+    , _filelockState(std::move(filelockState))
     , _privateLinkUrl(accountState->account()->deprecatedPrivateLinkUrl(numericFileId).toString(QUrl::FullyEncoded))
     , _startPage(startPage)
-    , _linkWidgetList({})
-    , _emptyShareLinkWidget(nullptr)
-    , _userGroupWidget(nullptr)
-    , _progressIndicator(nullptr)
 {
     setWindowFlags(windowFlags() & ~Qt::WindowContextHelpButtonHint);
     setAttribute(Qt::WA_DeleteOnClose);
@@ -76,18 +92,27 @@ ShareDialog::ShareDialog(QPointer<AccountState> accountState,
     }
 
     // Set filename
-    QFileInfo lPath(_localPath);
-    QString fileName = lPath.fileName();
+    QString fileName = QFileInfo(_sharePath).fileName();
     _ui->label_name->setText(tr("%1").arg(fileName));
     QFont f(_ui->label_name->font());
-    f.setPointSize(f.pointSize() * 1.4);
+    f.setPointSize(qRound(f.pointSize() * 1.4));
     _ui->label_name->setFont(f);
 
+    if (_filelockState._locked) {
+        static constexpr auto SECONDS_PER_MINUTE = 60;
+        const auto lockExpirationTime = _filelockState._lockTime + _filelockState._lockTimeout;
+        const auto remainingTime = QDateTime::currentDateTime().secsTo(QDateTime::fromSecsSinceEpoch(lockExpirationTime));
+        const auto remainingTimeInMinute = static_cast<int>(remainingTime > 0 ? remainingTime / SECONDS_PER_MINUTE : 0);
+        _ui->label_lockinfo->setText(tr("Locked by %1 - Expires in %2 minutes", "remaining time before lock expires", remainingTimeInMinute).arg(_filelockState._lockOwnerDisplayName).arg(remainingTimeInMinute));
+    } else {
+        _ui->label_lockinfo->setVisible(false);
+    }
+
     QString ocDir(_sharePath);
     ocDir.truncate(ocDir.length() - fileName.length());
 
-    ocDir.replace(QRegExp("^/*"), "");
-    ocDir.replace(QRegExp("/*$"), "");
+    ocDir.replace(QRegularExpression("^/*"), "");
+    ocDir.replace(QRegularExpression("/*$"), "");
 
     // Laying this out is complex because sharePath
     // may be in use or not.
@@ -109,7 +134,7 @@ ShareDialog::ShareDialog(QPointer<AccountState> accountState,
     }
 
     if (QFileInfo(_localPath).isFile()) {
-        ThumbnailJob *job = new ThumbnailJob(_sharePath, _accountState->account(), this);
+        auto *job = new ThumbnailJob(_sharePath, _accountState->account(), this);
         connect(job, &ThumbnailJob::jobFinished, this, &ShareDialog::slotThumbnailFetched);
         job->start();
     }
@@ -125,79 +150,98 @@ ShareDialog::ShareDialog(QPointer<AccountState> accountState,
     connect(job, &PropfindJob::finishedWithError, this, &ShareDialog::slotPropfindError);
     job->start();
 
-    bool sharingPossible = true;
-    if (!accountState->account()->capabilities().sharePublicLink()) {
-        qCWarning(lcSharing) << "Link shares have been disabled";
-        sharingPossible = false;
-    } else if (!(maxSharingPermissions & SharePermissionShare)) {
-        qCWarning(lcSharing) << "The file can not be shared because it was shared without sharing permission.";
-        sharingPossible = false;
-    }
+    initShareManager();
+    
+    _scrollAreaViewPort = new QWidget(_ui->scrollArea);
+    _scrollAreaLayout = new QVBoxLayout(_scrollAreaViewPort);
+    _scrollAreaLayout->setContentsMargins(0, 0, 0, 0);
+    _ui->scrollArea->setWidget(_scrollAreaViewPort);
 
-    if (sharingPossible) {
-        _manager = new ShareManager(accountState->account(), this);
-        connect(_manager, &ShareManager::sharesFetched, this, &ShareDialog::slotSharesFetched);
-        connect(_manager, &ShareManager::linkShareCreated, this, &ShareDialog::slotAddLinkShareWidget);
-        connect(_manager, &ShareManager::linkShareRequiresPassword, this, &ShareDialog::slotLinkShareRequiresPassword);
-    }
+    _internalLinkWidget = new InternalLinkWidget(localPath, this);
+    _ui->verticalLayout->addWidget(_internalLinkWidget);
+    _internalLinkWidget->setupUiOptions();
+    connect(this, &ShareDialog::styleChanged, _internalLinkWidget, &InternalLinkWidget::slotStyleChanged);
+
+    adjustScrollWidget();
 }
 
-void ShareDialog::addLinkShareWidget(const QSharedPointer<LinkShare> &linkShare){
-    _linkWidgetList.append(new ShareLinkWidget(_accountState->account(), _sharePath, _localPath, _maxSharingPermissions, this));
-    int index = _linkWidgetList.size()-1;
-    _linkWidgetList.at(index)->setLinkShare(linkShare);
+ShareLinkWidget *ShareDialog::addLinkShareWidget(const QSharedPointer<LinkShare> &linkShare)
+{
+    const auto linkShareWidget = new ShareLinkWidget(_accountState->account(), _sharePath, _localPath, _maxSharingPermissions, _ui->scrollArea);
+    _linkWidgetList.append(linkShareWidget);
 
-    connect(linkShare.data(), &Share::serverError, _linkWidgetList.at(index), &ShareLinkWidget::slotServerError);
-    connect(linkShare.data(), &Share::shareDeleted, _linkWidgetList.at(index), &ShareLinkWidget::slotDeleteShareFetched);
-    connect(_manager, &ShareManager::linkShareRequiresPassword, _linkWidgetList.at(index), &ShareLinkWidget::slotCreateShareRequiresPassword);
-    connect(_manager, &ShareManager::serverError, _linkWidgetList.at(index), &ShareLinkWidget::slotServerError);
+    linkShareWidget->setLinkShare(linkShare);
+
+    connect(linkShare.data(), &Share::serverError, linkShareWidget, &ShareLinkWidget::slotServerError);
+    connect(linkShare.data(), &Share::shareDeleted, linkShareWidget, &ShareLinkWidget::slotDeleteShareFetched);
+
+    if(_manager) {
+        connect(_manager, &ShareManager::serverError, linkShareWidget, &ShareLinkWidget::slotServerError);
+    }
 
     // Connect all shares signals to gui slots
-    connect(this, &ShareDialog::toggleAnimation, _linkWidgetList.at(index), &ShareLinkWidget::slotToggleAnimation);
-    connect(_linkWidgetList.at(index), &ShareLinkWidget::createLinkShare, this, &ShareDialog::slotCreateLinkShare);
-    connect(_linkWidgetList.at(index), &ShareLinkWidget::deleteLinkShare, this, &ShareDialog::slotDeleteShare);
-    //connect(_linkWidgetList.at(index), &ShareLinkWidget::resizeRequested, this, &ShareDialog::slotAdjustScrollWidgetSize);
-
+    connect(this, &ShareDialog::toggleShareLinkAnimation, linkShareWidget, &ShareLinkWidget::slotToggleShareLinkAnimation);
+    connect(linkShareWidget, &ShareLinkWidget::createLinkShare, this, &ShareDialog::slotCreateLinkShare);
+    connect(linkShareWidget, &ShareLinkWidget::deleteLinkShare, this, &ShareDialog::slotDeleteShare);
+    connect(linkShareWidget, &ShareLinkWidget::createPassword, this, &ShareDialog::slotCreatePasswordForLinkShare);
+    
     // Connect styleChanged events to our widget, so it can adapt (Dark-/Light-Mode switching)
-    connect(this, &ShareDialog::styleChanged, _linkWidgetList.at(index), &ShareLinkWidget::slotStyleChanged);
+    connect(this, &ShareDialog::styleChanged, linkShareWidget, &ShareLinkWidget::slotStyleChanged);
 
-    _ui->verticalLayout->insertWidget(_linkWidgetList.size()+1, _linkWidgetList.at(index));
-    _linkWidgetList.at(index)->setupUiOptions();
+    _ui->verticalLayout->insertWidget(_linkWidgetList.size() + 1, linkShareWidget);
+    _scrollAreaLayout->addWidget(linkShareWidget);
+    
+    linkShareWidget->setupUiOptions();
+    adjustScrollWidget();
+
+    return linkShareWidget;
 }
 
-void ShareDialog::initLinkShareWidget(){
-    if(_linkWidgetList.size() == 0){
-        _emptyShareLinkWidget = new ShareLinkWidget(_accountState->account(), _sharePath, _localPath, _maxSharingPermissions, this);
+void ShareDialog::initLinkShareWidget()
+{
+    if(_linkWidgetList.size() == 0) {
+        _emptyShareLinkWidget = new ShareLinkWidget(_accountState->account(), _sharePath, _localPath, _maxSharingPermissions, _ui->scrollArea);
         _linkWidgetList.append(_emptyShareLinkWidget);
 
-//        connect(_emptyShareLinkWidget, &ShareLinkWidget::resizeRequested, this, &ShareDialog::slotAdjustScrollWidgetSize);
-//        connect(this, &ShareDialog::toggleAnimation, _emptyShareLinkWidget, &ShareLinkWidget::slotToggleAnimation);
+        _emptyShareLinkWidget->slotStyleChanged(); // Get the initial customizeStyle() to happen
+
+        connect(this, &ShareDialog::toggleShareLinkAnimation, _emptyShareLinkWidget, &ShareLinkWidget::slotToggleShareLinkAnimation);
+        connect(this, &ShareDialog::styleChanged, _emptyShareLinkWidget, &ShareLinkWidget::slotStyleChanged);
+
         connect(_emptyShareLinkWidget, &ShareLinkWidget::createLinkShare, this, &ShareDialog::slotCreateLinkShare);
+        connect(_emptyShareLinkWidget, &ShareLinkWidget::createPassword, this, &ShareDialog::slotCreatePasswordForLinkShare);
 
         _ui->verticalLayout->insertWidget(_linkWidgetList.size()+1, _emptyShareLinkWidget);
+        _scrollAreaLayout->addWidget(_emptyShareLinkWidget);
         _emptyShareLinkWidget->show();
-
-    } else if(_emptyShareLinkWidget) {
+    } else if (_emptyShareLinkWidget) {
         _emptyShareLinkWidget->hide();
         _ui->verticalLayout->removeWidget(_emptyShareLinkWidget);
         _linkWidgetList.removeAll(_emptyShareLinkWidget);
         _emptyShareLinkWidget = nullptr;
     }
+
+    adjustScrollWidget();
 }
 
-void ShareDialog::slotAddLinkShareWidget(const QSharedPointer<LinkShare> &linkShare){
-    emit toggleAnimation(true);
-    addLinkShareWidget(linkShare);
+void ShareDialog::slotAddLinkShareWidget(const QSharedPointer<LinkShare> &linkShare)
+{
+    emit toggleShareLinkAnimation(true);
+    const auto addedLinkShareWidget = addLinkShareWidget(linkShare);
     initLinkShareWidget();
-    emit toggleAnimation(false);
+    if (linkShare->isPasswordSet()) {
+        addedLinkShareWidget->focusPasswordLineEdit();
+    }
+    emit toggleShareLinkAnimation(false);
 }
 
 void ShareDialog::slotSharesFetched(const QList<QSharedPointer<Share>> &shares)
 {
-    emit toggleAnimation(true);
+    emit toggleShareLinkAnimation(true);
 
     const QString versionString = _accountState->account()->serverVersion();
     qCInfo(lcSharing) << versionString << "Fetched" << shares.count() << "shares";
+    
     foreach (auto share, shares) {
         if (share->getShareType() != Share::TypeLink || share->getUidOwner() != share->account()->davUser()) {
             continue;
@@ -208,18 +252,29 @@ void ShareDialog::slotSharesFetched(const QList<QSharedPointer<Share>> &shares)
     }
 
     initLinkShareWidget();
-    emit toggleAnimation(false);
+    emit toggleShareLinkAnimation(false);
 }
 
-// TODO
-void ShareDialog::slotAdjustScrollWidgetSize()
+void ShareDialog::adjustScrollWidget()
 {
-    int count = this->findChildren<ShareLinkWidget *>().count();
-    _ui->scrollArea->setVisible(count > 0);
-    if (count > 0 && count <= 3) {
-        _ui->scrollArea->setFixedHeight(_ui->scrollArea->widget()->sizeHint().height());
+    _ui->scrollArea->setVisible(_scrollAreaLayout->count() > 0);
+
+    // Sometimes the contentRect returns a height of 0, so we need a backup plan
+    const auto scrollAreaContentHeight = _scrollAreaLayout->contentsRect().height();
+
+    auto linkWidgetHeights = 0;
+
+    if(scrollAreaContentHeight == 0 && !_linkWidgetList.empty()) {
+        for (const auto linkWidget : _linkWidgetList) {
+            linkWidgetHeights += linkWidget->height() - 10;
+        }
     }
-    _ui->scrollArea->setFrameShape(count > 3 ? QFrame::StyledPanel : QFrame::NoFrame);
+
+    const auto overAvailableHeight = scrollAreaContentHeight > _ui->scrollArea->height() ||
+            linkWidgetHeights > _ui->scrollArea->height();
+
+    _ui->scrollArea->setFrameShape(overAvailableHeight ? QFrame::StyledPanel : QFrame::NoFrame);
+    _ui->verticalLayout->setSpacing(overAvailableHeight ? 10 : 0);
 }
 
 ShareDialog::~ShareDialog()
@@ -274,54 +329,104 @@ void ShareDialog::showSharingUi()
 
     if (!canReshare) {
         auto label = new QLabel(this);
-        label->setText(tr("The file can not be shared because it was shared without sharing permission."));
+        label->setText(tr("The file cannot be shared because it does not have sharing permission."));
         label->setWordWrap(true);
         _ui->verticalLayout->insertWidget(1, label);
         return;
     }
 
-    // We only do user/group sharing from 8.2.0
-    bool userGroupSharing =
-        theme->userGroupSharing()
-        && _accountState->account()->serverVersionInt() >= Account::makeServerVersion(8, 2, 0);
-
-    if (userGroupSharing) {
-        _userGroupWidget = new ShareUserGroupWidget(_accountState->account(), _sharePath, _localPath, _maxSharingPermissions, _privateLinkUrl, this);
-
+    if (theme->userGroupSharing()) {
+        _userGroupWidget = new ShareUserGroupWidget(_accountState->account(), _sharePath, _localPath, _maxSharingPermissions, _privateLinkUrl, _ui->scrollArea);
+        _userGroupWidget->getShares();
+        
         // Connect styleChanged events to our widget, so it can adapt (Dark-/Light-Mode switching)
         connect(this, &ShareDialog::styleChanged, _userGroupWidget, &ShareUserGroupWidget::slotStyleChanged);
 
+        _userGroupWidget->slotStyleChanged();
+
         _ui->verticalLayout->insertWidget(1, _userGroupWidget);
-        _userGroupWidget->getShares();
+        _scrollAreaLayout->addLayout(_userGroupWidget->shareUserGroupLayout());
     }
 
+    initShareManager();
+
     if (theme->linkSharing()) {
-        _manager->fetchShares(_sharePath);
+        if(_manager) {
+            _manager->fetchShares(_sharePath);
+        }
+    }
+
+    adjustScrollWidget();
+}
+
+void ShareDialog::initShareManager()
+{
+    bool sharingPossible = true;
+    if (!_accountState->account()->capabilities().sharePublicLink()) {
+        qCWarning(lcSharing) << "Link shares have been disabled";
+        sharingPossible = false;
+    } else if (!(_maxSharingPermissions & SharePermissionShare)) {
+        qCWarning(lcSharing) << "The file cannot be shared because it does not have sharing permission.";
+        sharingPossible = false;
+    }
+
+    if (!_manager && sharingPossible) {
+        _manager = new ShareManager(_accountState->account(), this);
+        connect(_manager, &ShareManager::sharesFetched, this, &ShareDialog::slotSharesFetched);
+        connect(_manager, &ShareManager::linkShareCreated, this, &ShareDialog::slotAddLinkShareWidget);
+        connect(_manager, &ShareManager::linkShareRequiresPassword, this, &ShareDialog::slotLinkShareRequiresPassword);
     }
 }
 
 void ShareDialog::slotCreateLinkShare()
 {
-    _manager->createLinkShare(_sharePath, QString(), QString());
+    if(_manager) {
+        const auto askOptionalPassword = _accountState->account()->capabilities().sharePublicLinkAskOptionalPassword();
+        const auto password = askOptionalPassword ? createRandomPassword() : QString();
+        _manager->createLinkShare(_sharePath, QString(), password);
+    }
 }
 
-void ShareDialog::slotLinkShareRequiresPassword()
+void ShareDialog::slotCreatePasswordForLinkShare(const QString &password)
 {
-    bool ok;
-    QString password = QInputDialog::getText(this,
-                                             tr("Password for share required"),
-                                             tr("Please enter a password for your link share:"),
-                                             QLineEdit::Normal,
-                                             QString(),
-                                             &ok);
-
-    if (!ok) {
-        // The dialog was canceled so no need to do anything
-        return;
+    const auto shareLinkWidget = qobject_cast<ShareLinkWidget*>(sender());
+    Q_ASSERT(shareLinkWidget);
+    if (shareLinkWidget) {
+        connect(_manager, &ShareManager::linkShareRequiresPassword, shareLinkWidget, &ShareLinkWidget::slotCreateShareRequiresPassword);
+        connect(shareLinkWidget, &ShareLinkWidget::createPasswordProcessed, this, &ShareDialog::slotCreatePasswordForLinkShareProcessed);
+        shareLinkWidget->getLinkShare()->setPassword(password);
+    } else {
+        qCCritical(lcSharing) << "shareLinkWidget is not a sender!";
     }
+}
 
-    // Try to create the link share again with the newly entered password
-    _manager->createLinkShare(_sharePath, QString(), password);
+void ShareDialog::slotCreatePasswordForLinkShareProcessed()
+{
+    const auto shareLinkWidget = qobject_cast<ShareLinkWidget*>(sender());
+    Q_ASSERT(shareLinkWidget);
+    if (shareLinkWidget) {
+        disconnect(_manager, &ShareManager::linkShareRequiresPassword, shareLinkWidget, &ShareLinkWidget::slotCreateShareRequiresPassword);
+        disconnect(shareLinkWidget, &ShareLinkWidget::createPasswordProcessed, this, &ShareDialog::slotCreatePasswordForLinkShareProcessed);
+    } else {
+        qCCritical(lcSharing) << "shareLinkWidget is not a sender!";
+    }
+}
+
+void ShareDialog::slotLinkShareRequiresPassword(const QString &message)
+{
+    const auto passwordInputDialog = new PasswordInputDialog(tr("Please enter a password for your link share:"), message, this);
+    passwordInputDialog->setWindowTitle(tr("Password for share required"));
+    passwordInputDialog->setAttribute(Qt::WA_DeleteOnClose);
+    passwordInputDialog->open();
+
+    connect(passwordInputDialog, &QDialog::finished, this, [this, passwordInputDialog](const int result) {
+        if (result == QDialog::Accepted && _manager) {
+            // Try to create the link share again with the newly entered password
+            _manager->createLinkShare(_sharePath, QString(), passwordInputDialog->password());
+            return;
+        }
+        emit toggleShareLinkAnimation(false);
+    });
 }
 
 void ShareDialog::slotDeleteShare()
@@ -329,6 +434,7 @@ void ShareDialog::slotDeleteShare()
     auto sharelinkWidget = dynamic_cast<ShareLinkWidget*>(sender());
     sharelinkWidget->hide();
     _ui->verticalLayout->removeWidget(sharelinkWidget);
+    _scrollAreaLayout->removeWidget(sharelinkWidget);
     _linkWidgetList.removeAll(sharelinkWidget);
     initLinkShareWidget();
 }
@@ -352,7 +458,7 @@ void ShareDialog::slotAccountStateChanged(int state)
     bool enabled = (state == AccountState::State::Connected);
     qCDebug(lcSharing) << "Account connected?" << enabled;
 
-    if (_userGroupWidget != nullptr) {
+    if (_userGroupWidget) {
         _userGroupWidget->setEnabled(enabled);
     }
 
@@ -379,4 +485,10 @@ void ShareDialog::changeEvent(QEvent *e)
     QDialog::changeEvent(e);
 }
 
+void ShareDialog::resizeEvent(QResizeEvent *event)
+{
+    adjustScrollWidget();
+    QDialog::resizeEvent(event);
 }
+
+} // namespace OCC
