@@ -273,6 +273,7 @@ void PropagateItemJob::done(SyncFileItem::Status statusArg, const QString &error
     case SyncFileItem::BlacklistedError:
     case SyncFileItem::FileLocked:
     case SyncFileItem::FileNameInvalid:
+    case SyncFileItem::FileNameClash:
         // nothing
         break;
     }
@@ -482,6 +483,8 @@ void OwncloudPropagator::start(SyncFileItemVector &&items)
 {
     Q_ASSERT(std::is_sorted(items.begin(), items.end()));
 
+    _abortRequested = false;
+
     /* This builds all the jobs needed for the propagation.
      * Each directory is a PropagateDirectory job, which contains the files in it.
      * In order to do that we loop over the items. (which are sorted by destination)
@@ -682,7 +685,6 @@ bool OwncloudPropagator::localFileNameClash(const QString &relFile)
     Q_ASSERT(!file.isEmpty());
 
     if (!file.isEmpty() && Utility::fsCasePreserving()) {
-        qCDebug(lcPropagator) << "CaseClashCheck for " << file;
 #ifdef Q_OS_MAC
         const QFileInfo fileInfo(file);
         if (!fileInfo.exists()) {
@@ -718,6 +720,7 @@ bool OwncloudPropagator::localFileNameClash(const QString &relFile)
         const QString fn = fileInfo.fileName();
         const QStringList list = fileInfo.dir().entryList({ fn });
         if (list.count() > 1 || (list.count() == 1 && list[0] != fn)) {
+            qCWarning(lcPropagator) << "Detected case clash between" << file << "and" << list.constFirst();
             return true;
         }
 #endif
@@ -841,7 +844,7 @@ bool OwncloudPropagator::createConflict(const SyncFileItemPtr &item,
     QString renameError;
     auto conflictModTime = FileSystem::getModTime(fn);
     if (conflictModTime <= 0) {
-        *error = tr("Impossible to get modification time for file in conflict %1)").arg(fn);
+        *error = tr("Impossible to get modification time for file in conflict %1").arg(fn);
         return false;
     }
     QString conflictUserName;
@@ -1201,7 +1204,7 @@ void PropagateDirectory::slotSubJobsFinished(SyncFileItem::Status status)
 
             if (_item->_modtime <= 0) {
                 status = _item->_status = SyncFileItem::NormalError;
-                _item->_errorString = tr("Error updating metadata due to invalid modified time");
+                _item->_errorString = tr("Error updating metadata due to invalid modification time");
                 qCWarning(lcDirectory) << "Error writing to the database for file" << _item->_file;
             }
 
